@@ -171,6 +171,13 @@ namespace dlib
             ~tensor_conv (
             );
 
+            void operator() (
+                const bool add_to_output,
+                resizable_tensor& output,
+                const tensor& data,
+                const tensor& filters
+            );
+
             void operator() (
                 const bool add_to_output,
                 tensor& output,
@@ -182,7 +189,16 @@ namespace dlib
                 const bool add_to_output,
                 resizable_tensor& output,
                 const tensor& data,
-                const tensor& filters
+                const tensor& filters,
+                const tensor& biases
+            );
+
+            void operator() (
+                const bool add_to_output,
+                tensor& output,
+                const tensor& data,
+                const tensor& filters,
+                const tensor& biases
             );
 
             void get_gradient_for_data (
@@ -208,6 +224,16 @@ namespace dlib
                 int padding_x
             );
 
+           void setup(
+                const tensor& data,
+                const tensor& filters,
+                const tensor& biases,
+                int stride_y,
+                int stride_x,
+                int padding_y,
+                int padding_x
+            );
+
         private:
 
             // These variables record the type of data given to the last call to setup().
@@ -382,7 +408,7 @@ namespace dlib
 
     // ------------------------------------------------------------------------------------
 
-    } 
+    }
 }
 
 #endif // DLIB_USE_CUDA
